# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import multiprocessing
from pyworkflow.protocol.params import PointerParam, FileParam
from pwem.protocols import BatchProtocol
from pwem.objects import SetOfClasses2D
from xmipp3.convert import writeSetOfParticles, writeSetOfVolumes, readSetOfVolumes

from pyworkflow.utils import runCommand
from pwem.emlib.image import ImageHandler
import pwem.emlib.metadata as md


class FlexBatchProtClusterSet(BatchProtocol):
    """ Protocol executed when a set of cluster is created
    from set of pdbs.
    """
    _label = 'cluster set'

    def _defineParams(self, form):
        form.addHidden('inputSet', PointerParam, pointerClass='SetOfClasses2D,SetOfClasses3D')
        form.addHidden('inputSet', PointerParam, pointerClass='SetOfClasses2D,SetOfClasses3D')
        form.addParallelSection(threads=1, mpi=multiprocessing.cpu_count()//2-1)

    # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        self._insertFunctionStep('convertInputStep')
        self._insertFunctionStep('reconstructStep')
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------

    def convertInputStep(self):
        pass

    def reconstructStep(self):
        inputClasses = self.inputSet.get()

        for i in inputClasses:
            if i.getObjId() != 0:
                classFile = self._getExtraPath("class%i.xmd" % i.getObjId())
                if isinstance(inputClasses, SetOfClasses2D):
                    writeSetOfParticles(i, classFile)
                else:
                    writeSetOfVolumes(i,classFile)

        for i in inputClasses:
            if i.getObjId() != 0:
                classFile = self._getExtraPath("class%i.xmd" % i.getObjId())
                classVol = self._getExtraPath("class%i.vol" % i.getObjId())
                if isinstance(inputClasses, SetOfClasses2D):
                    args = "-i %s -o %s " % (classFile, classVol)
                    if self.numberOfMpi.get() > 1 :
                        progname = "xmipp_mpi_reconstruct_fourier "
                        self.runJob(progname, args)
                    else:
                        progname = "xmipp_reconstruct_fourier "
                        runCommand(progname + args)
                else:
                    classAvg = ImageHandler().computeAverage(i)
                    classAvg.write(classVol)

    def createOutputStep(self):
        outputMd = md.MetaData()
        inputClasses = self.inputSet.get()
        for i in inputClasses:
            if i.getObjId() != 0:
                classVol = self._getExtraPath("class%i.vol" % i.getObjId())
                index = outputMd.addObject()
                outputMd.setValue(md.MDL_IMAGE, classVol, index)
                outputMd.setValue(md.MDL_ITEM_ID, i.getObjId(), index)
        outputMd.write(self._getExtraPath("outputVols.xmd"))
        outputVols = self._createSetOfVolumes()
        readSetOfVolumes(self._getExtraPath("outputVols.xmd"),outputVols)
        outputVols.setSamplingRate(inputClasses.getSamplingRate())
        self._defineOutputs(outputVols=outputVols)
    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        return []



# **************************************************************************
# *
# * Authors:
# * Mohamad Harastani (mohamad.harastani@igbmc.fr)
# * J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# * Slavica Jonic (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from collections import OrderedDict
from pwem.emlib import (MDL_NMA_MODEFILE, MDL_NMA_COLLECTIVITY, MDL_NMA_SCORE, MDL_ORDER)
from pyworkflow.utils import Environ
from pwem.objects import NormalMode

from xmipp3.convert import rowToObject, objectToRow
from continuousflex.constants import NMA_HOME
import numpy as np
import math

            
MODE_DICT = OrderedDict([ 
       ("_modeFile", MDL_NMA_MODEFILE),
       ("_collectivity", MDL_NMA_COLLECTIVITY),
       ("_score", MDL_NMA_SCORE),
])


def rowToMode(row):
    """ Set properties of a NormalMode object from a Metadata row. """
    mode = NormalMode()
    rowToObject(row, mode, MODE_DICT)
    mode.setObjId(row.getValue(MDL_ORDER))
    return mode


def modeToRow(mode, row):
    """ Write the MetaData row from a given NormalMode object. """
    row.setValue(MDL_ORDER, int(mode.getObjId()))
    objectToRow(mode, row, MODE_DICT)
    
    
def getNMAEnviron():
    """ Create the needed environment for NMA programs. """
    from continuousflex import Plugin
    environ = Plugin.getEnviron()
    environ.update({'PATH': Plugin.getVar(NMA_HOME)}, position=Environ.BEGIN)
    environ.update({'LD_LIBRARY_PATH': Plugin.getCondaLibPath()}, position=Environ.BEGIN)
    return environ


def eulerAngles2matrix(alpha, beta, gamma, shiftx, shifty, shiftz):
    A = np.empty([4,4])
    A.fill(2)
    A[3,3] = 1
    A[3,0:3] = 0
    A[0,3] = float(shiftx)
    A[1,3] = float(shifty)
    A[2,3] = float(shiftz)
    alpha = float(alpha)
    beta = float(beta)
    gamma = float(gamma)
    sa = np.sin(np.deg2rad(alpha))
    ca = np.cos(np.deg2rad(alpha))
    sb = np.sin(np.deg2rad(beta))
    cb = np.cos(np.deg2rad(beta))
    sg = np.sin(np.deg2rad(gamma))
    cg = np.cos(np.deg2rad(gamma))
    cc = cb * ca
    cs = cb * sa
    sc = sb * ca
    ss = sb * sa
    A[0,0] = cg * cc - sg * sa
    A[0,1] = cg * cs + sg * ca
    A[0,2] = -cg * sb
    A[1,0] = -sg * cc - cg * sa
    A[1,1] = -sg * cs + cg * ca
    A[1,2] = sg * sb
    A[2,0] = sc
    A[2,1] = ss
    A[2,2] = cb
    return A


def matrix2eulerAngles(A):
    abs_sb = np.sqrt(A[0, 2] * A[0, 2] + A[1, 2] * A[1, 2])
    if (abs_sb > 16*np.exp(-5)):
        gamma = math.atan2(A[1, 2], -A[0, 2])
        alpha = math.atan2(A[2, 1], A[2, 0])
        if (abs(np.sin(gamma)) < np.exp(-5)):
            sign_sb = np.sign(-A[0, 2] / np.cos(gamma))
        else:
            if np.sin(gamma) > 0:
                sign_sb = np.sign(A[1, 2])
            else:
                sign_sb = -np.sign(A[1, 2])
        beta = math.atan2(sign_sb * abs_sb, A[2, 2])
    else:
        if (np.sign(A[2, 2]) > 0):
            alpha = 0
            beta  = 0
            gamma = math.atan2(-A[1, 0], A[0, 0])
        else:
            alpha = 0
            beta  = np.pi
            gamma = math.atan2(A[1, 0], -A[0, 0])
    gamma = np.rad2deg(gamma)
    beta  = np.rad2deg(beta)
    alpha = np.rad2deg(alpha)
    return alpha, beta, gamma, A[0,3], A[1,3], A[2,3]


def l2(Vec1, Vec2):
    Vec1 = np.array(Vec1)
    Vec2 = np.array(Vec2)
    value = np.inner(Vec1-Vec2, Vec1-Vec2)
    return np.sqrt(value)



# **************************************************************************
# *
# * Authors:     Ilyes Hamitouche (ilyes.hamitouche@upmc.fr)
# * IMPMC, Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pwem.protocols import ProtImportPdb, ProtImportParticles, ProtSubSet
from pwem.tests.workflows import TestWorkflow
from pyworkflow.tests import setupTestProject, DataSet
from continuousflex.protocols import (FlexProtNMA, FlexProtAlignmentNMA,
                                      FlexProtDimredNMA, NMA_CUTOFF_ABS,
                                      FlexProtDeepHEMNMATrain,
                                      FlexProtDeepHEMNMAInfer)
from xmipp3.protocols import XmippProtCropResizeParticles


class TestDeepHEMNMA(TestWorkflow):
    """ Test protocol for deepHEMNMA. """
    @classmethod
    def setUpClass(cls):    
        # Create a new project
        setupTestProject(cls)
        cls.ds = DataSet.getDataSet('nma_V2.0')
    
    def test_HEMNMA_atomic(self):
        protImportPdb = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                         pdbFile=self.ds.getFile('pdb'))
        protImportPdb.setObjLabel('AK.pdb')
        self.launchProtocol(protImportPdb)
        
        # Launch NMA for PDB imported
        protNMA1 = self.newProtocol(FlexProtNMA,
                                    cutoffMode=NMA_CUTOFF_ABS)
        protNMA1.inputStructure.set(protImportPdb.outputPdb)
        protNMA1.setObjLabel('NMA')
        self.launchProtocol(protNMA1)
        
        # Import the set of particles 
        # (in this order just to be in the middle in the tree)
        protImportParts = self.newProtocol(ProtImportParticles,
                                           filesPath=self.ds.getFile('particles'),
                                           samplingRate=1.0)
        protImportParts.setObjLabel('Particles')
        self.launchProtocol(protImportParts)

        protResizeParts= self.newProtocol(XmippProtCropResizeParticles)
        protResizeParts.doResize.set(True)
        protResizeParts.resizeOption.set(2) # this corresponds to factor
        protResizeParts.resizeFactor.set(0.25)
        protResizeParts.inputParticles.set(protImportParts.outputParticles)
        protResizeParts.setObjLabel('Resizing (factor 0.25)')
        self.launchProtocol(protResizeParts)

        protSubset1 = self.newProtocol(ProtSubSet,
                                      objLabel='Training set',
                                      chooseAtRandom=True,
                                      nElements=4)
        # protSubset1.inputFullSet.set(protImportParts.outputParticles)
        protSubset1.inputFullSet.set(protResizeParts.outputParticles)
        self.launchProtocol(protSubset1)


        protSubset2 = self.newProtocol(ProtSubSet,
                                         objLabel='Inference set',
                                         chooseAtRandom=False,
                                         setOperation=1)
        protSubset2.inputFullSet.set(protResizeParts.outputParticles)
        # protSubset2.inputFullSet.set(protImportParts.outputParticles)
        protSubset2.inputSubSet.set(protSubset1.outputParticles)
        self.launchProtocol(protSubset2)

        # Launch NMA alignment
        protAlignment = self.newProtocol(FlexProtAlignmentNMA,
                                         modeList='7-9')
        protAlignment.inputModes.set(protNMA1.outputModes)
        protAlignment.inputParticles.set(protSubset1.outputParticles)
        protAlignment.setObjLabel('HEMNMA atomic ref')
        self.launchProtocol(protAlignment)

        protTrain = self.newProtocol(FlexProtDeepHEMNMATrain)
        protTrain.inputNMA.set(protAlignment)
        self.launchProtocol(protTrain)

        protInfer = self.newProtocol(FlexProtDeepHEMNMAInfer)
        protInfer.trained_model.set(protTrain) #angles and shifts
        protInfer.inputParticles.set(protSubset2.outputParticles)
        self.launchProtocol(protInfer)



# Author: Ilyes Hamitouche

from torchvision import transforms
from torch.utils.data import DataLoader
from processing_dh.data import cryodata
from processing_dh.utils import quater2euler, reverse_min_max
from processing_dh.models import deephemnma
import numpy as np
import torch
from pathlib import Path
import sys
import pwem.emlib.metadata as md
def norm(imgs_path, weights_path, flag, mode, batch_size):
    dataset = cryodata(imgs_path, weights_path, flag=flag, mode=mode, transform=transforms.ToTensor())
    train_loader = DataLoader(dataset, batch_size=batch_size)
    sum_, squared_sum_, num_batches = 0, 0, 0
    for img, image_name in train_loader:
        sum_ += torch.mean(img, dim=[0, 2, 3])
        squared_sum_ += torch.mean(img**2, dim=[0, 2, 3])
        num_batches += 1
    mean = sum_/num_batches
    std = (squared_sum_/num_batches - mean**2)**0.5
    return mean, std
def infer(imgs_path, weights_path, output_path, num_modes, batch_size=2, flag=0, device=0, mode='inference'):
    FLAG = ''
    if flag==0:
        FLAG = 'nma'
    elif flag==1:
        FLAG = 'ang'
    elif flag==2:
        FLAG = 'shf'
    else:
        FLAG = 'all'
    DEVICE = ''
    if device==0:
        DEVICE = 'cuda'
    else:
        DEVICE = 'cpu'

    mean, std = norm(imgs_path, weights_path, flag=FLAG, mode=mode, batch_size=batch_size)
    transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((mean), (std))])
    dataset = cryodata(imgs_path, weights_path, flag=FLAG, mode=mode, transform=transform)

    dataset_size = len(dataset)
    print('the train set size is: {} images'.format(dataset_size))

    data_loader = DataLoader(dataset, batch_size=batch_size)

    if FLAG=='nma':
        model = deephemnma(3).to(DEVICE)
        predictions = np.zeros((dataset_size, num_modes), dtype='float32')
    elif FLAG=='ang':
        model = deephemnma(4).to(DEVICE)
        predictions = np.zeros((dataset_size, 4), dtype='float32')
    elif FLAG=='shf':
        model = deephemnma(2).to(DEVICE)
        predictions = np.zeros((dataset_size, 2), dtype='float32')
    elif FLAG=='all':
        model = deephemnma(6+num_modes).to(DEVICE)
        predictions = np.zeros((dataset_size, 6+num_modes), dtype='float32')

    model.load_state_dict(torch.load(weights_path))
    model.eval()
    with torch.no_grad():
        i = 0
        for img, image_name in data_loader:
            pred_params = model(img.to(DEVICE), mode)
            predictions[i * batch_size:(i + 1) * batch_size, :] = pred_params.cpu()
            i+=1


    if FLAG=='nma':
        min_max_nma = np.loadtxt(str(Path(weights_path).parent) + '/min_max_nma.txt')
        nma = reverse_min_max(predictions, min_max_nma[0], min_max_nma[1])
    elif FLAG=='ang':
        angles = predictions
        euler_angles = []
        for i in range(len(angles)):
            euler_angles.append(quater2euler(angles))
        euler_angles = np.array(euler_angles)
    elif FLAG=='shf':
        min_max_shf = np.loadtxt(str(Path(weights_path).parent) + '/min_max_shf.txt')
        shifts = reverse_min_max(predictions, min_max_shf[0], min_max_shf[1])
    elif FLAG=='all':
        min_max_nma = np.loadtxt(str(Path(weights_path).parent) + '/min_max_nma.txt')
        min_max_shf = np.loadtxt(str(Path(weights_path).parent) + '/min_max_shf.txt')
        nma = reverse_min_max(predictions[:,:num_modes], min_max_nma[0], min_max_nma[1])
        angles = predictions[:,num_modes:num_modes+4]
        shifts = reverse_min_max(predictions[:,num_modes+4:], min_max_shf[0], min_max_shf[1])
        euler_angles = []
        for i in range(len(angles)):
            euler_angles.append(quater2euler(angles[i]))
        euler_angles = np.array(euler_angles)
        mdImgs = md.MetaData(imgs_path)
        imgPath = []
        for objId in mdImgs:
            imgPath.append(mdImgs.getValue(md.MDL_IMAGE, objId))
        with open(output_path+'/images.xmd', 'w') as f:
            f.write('# XMIPP_STAR_1 * \n # \ndata_noname\nloop_\n _image\n _enabled\n _angleRot\n _angleTilt\n _anglePsi\n _shiftX\n _shiftY\n _nmaDisplacements\n _cost\n _itemId\n')
            for i in range(euler_angles.shape[0]):
                f.write(imgPath[i]+"                    1 {:>12.6} {:>12.6} {:>12.6} {:>12.6} {:>12.6} '{:>12.6} {:>12.6} {:>12.6}'   0.55  {}".format(
                    euler_angles[i, 0], euler_angles[i, 1], euler_angles[i, 2], shifts[i, 0],
                        shifts[i, 1], nma[i,0], nma[i,1], nma[i,2], i+1) + '\n')
if __name__ == '__main__':
    infer(sys.argv[1],
          sys.argv[2],
          sys.argv[3],
          int(sys.argv[4]),
          int(sys.argv[5]),
          int(sys.argv[6]),
          int(sys.argv[7]))



# **************************************************************************
# *
# * Authors:
# * Mohamad Harastani (mohamad.harastani@igbmc.fr)
# * J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# * Slavica Jonic (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import isfile
from pyworkflow.protocol.params import PointerParam, FileParam
from pwem.protocols import BatchProtocol
from pwem.objects import SetOfParticles, Volume, AtomStruct
from xmipp3.convert import writeSetOfParticles
from pwem.utils import runProgram
import pwem.emlib.metadata as md
import numpy as np


class FlexBatchProtNMACluster(BatchProtocol):
    """ Protocol executed when a cluster is created
    from NMA images and theirs deformations.
    """
    _label = 'nma cluster'
    
    def _defineParams(self, form):
        form.addHidden('inputNmaDimred', PointerParam, pointerClass='EMObject')
        form.addHidden('sqliteFile', FileParam)
        
    #--------------------------- INSERT steps functions --------------------------------------------
        
    def _insertAllSteps(self):
        imagesMd = self._getExtraPath('images.xmd')
        outputVol = self._getExtraPath('reconstruction.vol')
        
        self._insertFunctionStep('convertInputStep', imagesMd)
        params = '-i %(imagesMd)s -o %(outputVol)s --fast' % locals()
        self._insertFunctionStep('reconstructStep', params)
        self._insertFunctionStep('centroidPdbStep')
        self._insertFunctionStep('createOutputStep', outputVol)
        
    #--------------------------- STEPS functions --------------------------------------------   
        
    def convertInputStep(self, imagesMd):
        # It is unusual to create the output in the convertInputStep,
        # but just to avoid reading twice the sqlite with particles
        inputSet = self.inputNmaDimred.get().getInputParticles()
        partSet = self._createSetOfParticles()
        partSet.copyInfo(inputSet)
        
        tmpSet = SetOfParticles(filename=self.sqliteFile.get())        
        partSet.appendFromImages(tmpSet)
        # Register outputs
        partSet.setAlignmentProj()
        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(inputSet, partSet)
        
        writeSetOfParticles(partSet, imagesMd)

        # Add the NMA displacement to clusters XMD files
        md_file_nma = md.MetaData(self.inputNmaDimred.get().getParticlesMD())
        md_file_org = md.MetaData(imagesMd)
        for objID in md_file_org:
            # if image name is the same, we add the nma displacement from nma to org
            id_org = md_file_org.getValue(md.MDL_ITEM_ID, objID)
            for j in md_file_nma:
                id_nma = md_file_nma.getValue(md.MDL_ITEM_ID, j)
                print(id_nma)
                if id_org == id_nma:
                    displacements = md_file_nma.getValue(md.MDL_NMA, j)
                    md_file_org.setValue(md.MDL_NMA, displacements, objID)
                    break
        md_file_org.write(imagesMd)


    def reconstructStep(self, params):
        runProgram('xmipp_reconstruct_fourier_accel', params)


    def centroidPdbStep(self):
        imagesMd = self._getExtraPath('images.xmd')
        md_file = md.MetaData(imagesMd)
        deformations = []
        for j in md_file:
            defor = md_file.getValue(md.MDL_NMA, j)
            if defor:
                deformations.append(defor)
        ampl = np.mean(np.array(deformations), axis= 0)
        print(self.getFnPDB())

        fnPDB, pseudo = self.getFnPDB()
        fnModeList = self.getFnModes()
        fnOutPDB = self._getExtraPath('centroid.pdb')
        params = " --pdb " + fnPDB
        params += " --nma " + fnModeList
        params += " -o " + fnOutPDB
        params += " --deformations " + ' '.join(str(i) for i in ampl)
        runProgram('xmipp_pdb_nma_deform', params)
    
    def createOutputStep(self, outputVol):
        vol = Volume()
        vol.setFileName(outputVol)
        vol.setSamplingRate(self.outputParticles.getSamplingRate())
        atm = AtomStruct()
        fnPDB, pseudo = self.getFnPDB()
        fnOutPDB = self._getExtraPath('centroid.pdb')
        atm.setPseudoAtoms(pseudo)
        atm.setFileName(fnOutPDB)
        atm.setVolume(vol)
        self._defineOutputs(centroidPDB=atm)
        self._defineOutputs(outputVol=vol)

    #--------------------------- Utility functions -----------------------------------------
    def getFnPDB(self):
        # This functions returns the path of the structure, false if is atomic, true if pseudoatomic
        path = self.inputNmaDimred.get().inputNMA.get()._getExtraPath('atoms.pdb')
        if isfile(path):
            return path, False
        else:
            path = self.inputNmaDimred.get().inputNMA.get()._getExtraPath('pseudoatoms.pdb')
            return path, True

    def getFnModes(self):
        return self.inputNmaDimred.get().getInputModes()

    #--------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary
    
    def _validate(self):
        errors = []
        return errors
    
    def _citations(self):
        return ['harastani2022continuousflex']
    
    def _methods(self):
        return []
    



# **************************************************************************
# * Authors:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# *           Remi Vuillemot             (remi.vuillemot@upmc.fr)
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pyworkflow.protocol.params import (PointerParam, EnumParam, IntParam)
from pwem.protocols import ProtAnalysis3D
from pyworkflow.protocol import params
from continuousflex.protocols.utilities.genesis_utilities import numpyArr2dcd, dcd2numpyArr
from .utilities.pdb_handler import ContinuousFlexPDBHandler
from pwem.objects import AtomStruct, SetOfParticles, SetOfVolumes
from xmipp3.convert import writeSetOfVolumes, writeSetOfParticles, readSetOfVolumes, readSetOfParticles
from pwem.constants import ALIGN_PROJ
from continuousflex.protocols.convert import matrix2eulerAngles

import numpy as np
import glob
import pwem.emlib.metadata as md

PDB_SOURCE_PATTERN = 0
PDB_SOURCE_OBJECT = 1
PDB_SOURCE_TRAJECT = 2


class FlexProtAlignPdb(ProtAnalysis3D):
    """ Protocol to perform rigid body alignement on a set of PDB files. """
    _label = 'pdbs rigid body alignement'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('pdbSource', EnumParam, default=0,
                      label='Source of PDBs',
                      choices=['File pattern', 'Object', 'Trajectory Files'],
                      help='Use the file pattern as file location with /*.pdb')
        form.addParam('pdbs_file', params.PathParam,
                      condition='pdbSource == %i'%PDB_SOURCE_PATTERN,
                      label="List of PDBs",
                      help='Use the file pattern as file location with /*.pdb')
        form.addParam('setOfPDBs', params.PointerParam, pointerClass='SetOfPDBs, SetOfAtomStructs',
                      condition='pdbSource == %i'%PDB_SOURCE_OBJECT,
                      label="Set of PDBs",
                      help='Use a scipion object SetOfPDBs / SetOfAtomStructs')
        form.addParam('dcds_file', params.PathParam,
                      condition='pdbSource == %i'%PDB_SOURCE_TRAJECT,
                      label="DCD trajectory file (s)",
                      help='Use the file pattern as file location with /*.dcd')
        form.addParam('dcd_ref_pdb', params.PointerParam, pointerClass='AtomStruct',
                      condition='pdbSource == %i'%PDB_SOURCE_TRAJECT,
                      label="trajectory Reference PDB",
                      help='Reference PDB of the trajectory (Only used for structural information (Atom name, residue number etc)'
                           '. The coordinates inside this PDB are not used. The atoms number and position in the file must'
                           ' correspond to the DCD file. ')
        form.addParam('dcd_start', params.IntParam, default=0,
                      condition='pdbSource == %i'%PDB_SOURCE_TRAJECT,
                      label="Beginning of the trajectory",
                      help='Index of the desired begining of the trajectory', expertLevel=params.LEVEL_ADVANCED)
        form.addParam('dcd_end', params.IntParam, default=-1,
                      condition='pdbSource == %i'%PDB_SOURCE_TRAJECT,
                      label="Ending of the trajectory",
                      help='Index of the desired end of the trajectory', expertLevel=params.LEVEL_ADVANCED)
        form.addParam('dcd_step', params.IntParam, default=1,
                      condition='pdbSource == %i'%PDB_SOURCE_TRAJECT,
                      label="Step of the trajectory",
                      help='Step to skip points in the trajectory', expertLevel=params.LEVEL_ADVANCED)



        form.addParam('alignRefPDB', params.PointerParam, pointerClass='AtomStruct',
                      label="Alignement Reference PDB",
                      help='Reference PDB to align the PDBs with')
        form.addParam('matchingType', params.EnumParam, label="Match PDBs and reference PDB ?", default=0,
                      choices=['All PDBs are matching', 'Match chain name + residue no',
                               'Match segment name + residue no'],
                      help="Method to find atomic coordinates correspondence between the pdb set "
                           "coordinates and the reference PDB. The method will select the matching atoms"
                           " and sort them in the corresponding order. If the structures in the files are"
                           " already matching, choose All structures are matching")

        form.addParam('createOutput', params.BooleanParam, default=True,
                      label="Create output Set of PDBs ?",
                      help='Create output set. This step can be time consuming and not necessary if you are only '
                           ' interested by the alignement parameters. The aligned coordinate are conserved as DCD file '
                           'in the extra directory.'
                        , expertLevel=params.LEVEL_ADVANCED)

        form.addSection(label='Apply alignment to other set')
        form.addParam('applyAlignment', params.BooleanParam, default=False,
                      label="Apply alignment to other data set ?",
                      help='Use the PDB alignement to align another data set.')
        form.addParam('otherSet', params.PointerParam, pointerClass='SetOfParticles, SetOfVolumes',
                      condition='applyAlignment',
                      label="Other set of Particles / Volumes",
                      help='Use a scipion EMSet object')



        # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('readInputFiles')
        self._insertFunctionStep('rigidBodyAlignementStep')
        if self.applyAlignment.get():
            self._insertFunctionStep('applyAlignmentStep')
        if self.createOutput.get():
            self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def readInputFiles(self):
        inputFiles = self.getInputFiles()

        # Get pdbs coordinates
        if self.pdbSource.get() == PDB_SOURCE_TRAJECT:
            pdbs_arr = dcd2numpyArr(inputFiles[0])
            nframe, natom, _ = pdbs_arr.shape
            pdbs_arr = pdbs_arr
            for i in range(1,len(inputFiles)):
                pdb_arr_i = dcd2numpyArr(inputFiles[i])
                pdbs_arr = np.concatenate((pdbs_arr, pdb_arr_i), axis=0)

        else:
            pdbs_matrix = []
            for pdbfn in inputFiles:
                try:
                    # Read PDBs
                    mol = ContinuousFlexPDBHandler(pdbfn)
                    pdbs_matrix.append(mol.coords)
                except RuntimeError:
                    print("Warning : Can not read PDB file %s " % pdbfn)
            pdbs_arr = np.array(pdbs_matrix)

        # save as dcd file
        numpyArr2dcd(pdbs_arr, self._getExtraPath("coords.dcd"))

    def rigidBodyAlignementStep(self):

        # open files
        inputPDB = ContinuousFlexPDBHandler(self.getPDBRef())
        refPDB = ContinuousFlexPDBHandler(self.alignRefPDB.get().getFileName())
        arrDCD = dcd2numpyArr(self._getExtraPath("coords.dcd"))
        nframe, natom,_ =arrDCD.shape
        alignXMD = md.MetaData()

        # find matching index between reference and pdbs
        if self.matchingType.get() == 1:
            idx_matching_atoms = inputPDB.matchPDBatoms(reference_pdb=refPDB, matchingType=0)
            refPDB.select_atoms(idx_matching_atoms[:, 1])
        elif self.matchingType.get() == 2:
            idx_matching_atoms = inputPDB.matchPDBatoms(reference_pdb=refPDB, matchingType=1)
            refPDB.select_atoms(idx_matching_atoms[:, 1])
        else:
            idx_matching_atoms = None
        refPDB.write_pdb(self._getExtraPath("reference.pdb"))

        # loop over all pdbs
        for i in range(nframe):
            print("Aligning PDB %i ... " %i)

            # rotate
            if self.matchingType.get() != 0 :
                coord = arrDCD[i][idx_matching_atoms[:, 0]]
            else:
                coord = arrDCD[i]
            rot_mat, tran = ContinuousFlexPDBHandler.alignCoords(refPDB.coords, coord)
            arrDCD[i] = (np.dot(arrDCD[i], rot_mat) + tran).astype(np.float32)

            # add to MD
            trans_mat = np.zeros((4,4))
            trans_mat[:3,:3] = rot_mat
            trans_mat[:,3][:3] = tran
            rot, tilt, psi,shftx, shfty, shftz = matrix2eulerAngles(trans_mat)
            index = alignXMD.addObject()
            alignXMD.setValue(md.MDL_ANGLE_ROT, rot, index)
            alignXMD.setValue(md.MDL_ANGLE_TILT, tilt, index)
            alignXMD.setValue(md.MDL_ANGLE_PSI, psi, index)
            alignXMD.setValue(md.MDL_SHIFT_X, shftx, index)
            alignXMD.setValue(md.MDL_SHIFT_Y, shfty, index)
            alignXMD.setValue(md.MDL_SHIFT_Z, shftz, index)
            alignXMD.setValue(md.MDL_IMAGE, "", index)

        numpyArr2dcd(arrDCD, self._getExtraPath("coords.dcd"))
        alignXMD.write(self._getExtraPath("alignement.xmd"))


    def createOutputStep(self):
        pdbset = self._createSetOfPDBs("outputPDBs")
        arrDCD = dcd2numpyArr(self._getExtraPath("coords.dcd"))
        refPDB = ContinuousFlexPDBHandler(self._getExtraPath("reference.pdb"))

        nframe, natom,_ = arrDCD.shape
        for i in range(nframe):
            filename = self._getExtraPath("output_%s.pdb" %str(i+1).zfill(6))
            refPDB.coords = arrDCD[i]
            refPDB.write_pdb(filename)
            pdb = AtomStruct(filename=filename)
            pdbset.append(pdb)

        self._defineOutputs(outputPDBs = pdbset)

    def applyAlignmentStep(self):
        inputSet = self.otherSet.get()

        if isinstance(inputSet, SetOfVolumes):
            inputAlignement = self._createSetOfVolumes("inputAlignement")
            readSetOfVolumes(self._getExtraPath("alignement.xmd"), inputAlignement)
            alignedSet = self._createSetOfVolumes("alignedSet")
        else:
            inputAlignement = self._createSetOfParticles("inputAlignement")
            alignedSet = self._createSetOfParticles("alignedSet")
            readSetOfParticles(self._getExtraPath("alignement.xmd"), inputAlignement)

        alignedSet.setSamplingRate(inputSet.getSamplingRate())
        alignedSet.setAlignment(ALIGN_PROJ)
        iter1 = inputSet.iterItems()
        iter2 = inputAlignement.iterItems()
        for i in range(inputSet.getSize()):
            p1 = iter1.__next__()
            p2 = iter2.__next__()
            r1 = p1.getTransform()
            r2 = p2.getTransform()
            rot = r2.getRotationMatrix()
            tran = np.array(r2.getShifts()) / inputSet.getSamplingRate()
            # middle = np.ones(3) * p1.getDim()[0]/2 * inputSet.getSamplingRate()
            # new_tran = np.dot(middle, rot) + tran
            new_trans = np.zeros((4, 4))
            new_trans[:3, 3] = tran
            new_trans[:3, :3] = rot
            new_trans[3, 3] = 1.0
            r1.composeTransform(new_trans)
            p1.setTransform(r1)
            alignedSet.append(p1)
        self._defineOutputs(alignedSet = alignedSet)

        if isinstance(inputSet, SetOfVolumes):
            writeSetOfVolumes(alignedSet, self._getExtraPath("alignedSet.xmd"))
        else:
            writeSetOfParticles(alignedSet, self._getExtraPath("alignedSet.xmd"))
    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd',
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def getInputFiles(self):
        if self.pdbSource.get()==PDB_SOURCE_PATTERN:
            l= [f for f in glob.glob(self.pdbs_file.get())]
        elif self.pdbSource.get()==PDB_SOURCE_OBJECT:
            l= [i.getFileName() for i in self.setOfPDBs.get()]
        elif self.pdbSource.get()==PDB_SOURCE_TRAJECT:
            l= [f for f in glob.glob(self.dcds_file.get())]
        l.sort()
        return l

    def getPDBRef(self):
        if self.pdbSource.get()==PDB_SOURCE_TRAJECT:
            return self.dcd_ref_pdb.get().getFileName()
        else:
            return self.getInputFiles()[0]



# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import os.path
import subprocess
import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pwem.objects.data import AtomStruct, SetOfAtomStructs, SetOfPDBs, SetOfVolumes,SetOfParticles, Volume
import mrcfile
from pwem.utils import runProgram
from pyworkflow.utils import getListFromRangeString
from .utilities.genesis_utilities import *
from .utilities.pdb_handler import ContinuousFlexPDBHandler
import pyworkflow.utils as pwutils
from pyworkflow.utils import runCommand, buildRunCommand
from xmipp3.convert import writeSetOfParticles, writeSetOfVolumes
from pwem.convert.atom_struct import cifToPdb
from continuousflex import Plugin


class ProtGenesis(EMProtocol):
    """ Protocol to perform MD/NMMD simulation based on GENESIS. """
    _label = 'MD-NMMD-Genesis'

    def __init__(self, **kwargs):
        EMProtocol.__init__(self, **kwargs)
        self._inputEMMetadata =None

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):

        # Inputs ============================================================================================
        form.addSection(label='Inputs')

        form.addParam('inputType', params.EnumParam, label="Simulation inputs", default=INPUT_NEW_SIM,
                      choices=['New simulation from topology protocol', 'Restart previous GENESIS simulation', "New simulation from files"],
                      help="Chose the type of input for your simulation",
                      important=True)

        # INPUT_TOPOLOGY
        form.addParam('topoProt', params.PointerParam, label="Input topology protocol",
                      pointerClass="ProtGenerateTopology",
                       help='Provide a generate topology protocol to initialize the simulation',
                      condition="inputType==%i"%INPUT_TOPOLOGY)

        # INPUT_RESTART
        form.addParam('restartProt', params.PointerParam, label="Input GENESIS protocol",
                      pointerClass="ProtGenesis",
                       help='Provide a GENESIS protocol to restart.', condition="inputType==%i"%INPUT_RESTART)

        # INPUT_NEW_SIM
        form.addParam('inputPDB', params.PointerParam,
                      pointerClass='AtomStruct', label="Input PDB",
                      help='Select the input PDB.', important=True,  condition="inputType==%i"%INPUT_NEW_SIM)
        group = form.addGroup('Forcefield Inputs',  condition="inputType==%i"%INPUT_NEW_SIM)
        group.addParam('forcefield', params.EnumParam, label="Forcefield type", default=FORCEFIELD_CHARMM, important=True,

                      choices=['CHARMM', 'AAGO', 'CAGO'], help="Type of the force field used for energy and force calculation")

        group.addParam('inputTOP', params.FileParam, label="GROMACS Topology File (top)",
                      condition="(forcefield==%i or forcefield==%i)"%(FORCEFIELD_CAGO, FORCEFIELD_AAGO),
                      help='Gromacs ‘top’ file containing information of the system such as atomic masses, charges,'
                           ' atom connectivities. To generate this file for your system, you can either use the protocol'
                           '\" generate topology files\" (SMOG 2 installation is required, https://smog-server.org/smog2/ ),'
                           ' or using SMOG sever (https://smog-server.org/cgi-bin/GenTopGro.pl )')
        group.addParam('inputPRM', params.FileParam, label="CHARMM parameter file (prm)",
                      condition = "forcefield==%i"%FORCEFIELD_CHARMM,
                      help='CHARMM parameter file containing force field parameters, e.g. force constants and librium'
                            ' geometries. Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ' )
        group.addParam('inputRTF', params.FileParam, label="CHARMM topology file (rtf)",
                      condition="forcefield==%i"%FORCEFIELD_CHARMM,
                      help='CHARMM topology file containing information about atom connectivity of residues and'
                           ' other molecules. Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ')
        group.addParam('inputPSF', params.FileParam, label="CHARMM Structure File (psf)",
                      condition="forcefield==%i"%FORCEFIELD_CHARMM,
                      help='CHARMM/X-PLOR psf file containing information of the system such as atomic masses,'
                            ' charges, and atom connectivities. To generate this file for your system, you can either use the protocol'
                           '\" generate topology files\", VMD psfgen, or online CHARMM GUI ( https://www.charmm-gui.org/ ).')
        group.addParam('inputSTR', params.FileParam, label="CHARMM stream file (str, optional)",
                      condition="forcefield==%i"%FORCEFIELD_CHARMM, default="",
                      help='CHARMM stream file containing both topology information and parameters. '
                           'Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ')

        form.addParam('centerPDB', params.BooleanParam, label="Center PDB ?",
                      default=False, help="Center the input PDBs with the center of mass")

        group = form.addGroup('Execution Parameters',expertLevel=params.LEVEL_ADVANCED)
        group.addParam('disableParallelSim', params.BooleanParam, label="Disable parallelisation over the EM data ?", default=False,
                      help="Disabel parallel processing of simualtions over EM input data. Instead, each simulation is run linearly"
                           "with internal parallelization with the specified number of MPI. Running parallel simulation is activated by default "
                           " when using multiple EM data. Running parallel simulation is not available for REUS."
                           "",expertLevel=params.LEVEL_ADVANCED)
        group.addParam('raiseError', params.BooleanParam, label="Stop execution if fails ?", default=True,
                      help="Stop execution if GENESIS program fails",expertLevel=params.LEVEL_ADVANCED)
        group.addParam('md_program', params.EnumParam, label="MD program", default=PROGRAM_ATDYN,
                      choices=['ATDYN', 'SPDYN'],
                      help="SPDYN (Spatial decomposition dynamics) and ATDYN (Atomic decomposition dynamics)"
                    " share almost the same data structures, subroutines, and modules, but differ in"
                    " their parallelization schemes. In SPDYN, the spatial decomposition scheme is implemented with new"
                    " parallel algorithms and GPGPU calculation. In ATDYN, the atomic decomposition scheme"
                    " is introduced for simplicity. The performance of ATDYN is not comparable to SPDYN due to the"
                    " simple parallelization scheme but contains new methods and features. NMMD is available only for ATDYN.",
                      expertLevel=params.LEVEL_ADVANCED)

        # Simulation =================================================================================================
        form.addSection(label='Simulation')
        form.addParam('simulationType', params.EnumParam, label="Simulation type", default=0,
                      choices=['Minimization', 'Molecular Dynamics (MD)', 'Normal Mode Molecular Dynamics (NMMD)', 'Replica-Exchange MD', 'Replica-Exchange NMMD'],
                      help="Type of simulation to be performed by GENESIS", important=True)

        group = form.addGroup('Simulation parameters')
        group.addParam('integrator', params.EnumParam, label="Integrator", default=0,
                      choices=['Velocity Verlet', 'Leapfrog', ''],
                      help="Type of integrator for the simulation", condition="simulationType!=0")
        group.addParam('time_step', params.FloatParam, default=0.002, label='Time step (ps)',
                      help="Time step in the MD run", condition="simulationType!=0")
        group.addParam('n_steps', params.IntParam, default=10000, label='Number of steps',
                      help="Total number of steps in one MD run")
        group.addParam('eneout_period', params.IntParam, default=100, label='Energy output period',
                      help="Output frequency for the energy data")
        group.addParam('crdout_period', params.IntParam, default=100, label='Coordinate output period',
                      help="Output frequency for the coordinates data")
        group.addParam('nbupdate_period', params.IntParam, default=10, label='Non-bonded update period',
                      help="Update frequency of the non-bonded pairlist",
                      expertLevel=params.LEVEL_ADVANCED)

        group = form.addGroup('NMMD parameters', condition="simulationType==2 or simulationType==4")
        group.addParam('inputModes', params.PointerParam, pointerClass="SetOfNormalModes", label='Input Modes',
                       default=None,
                       help="Input set of normal modes", condition="simulationType==2 or simulationType==4")
        group.addParam('modeList', params.NumericRangeParam, expertLevel=params.LEVEL_ADVANCED,
                      label="Modes selection",
                      help='Select the normal modes that will be used for analysis. \n'
                           'If you leave this field empty, all computed modes will be selected for simulation.\n'
                           'You have several ways to specify the modes.\n'
                           '   Examples:\n'
                           ' "7,8-10" -> [7,8,9,10]\n'
                           ' "8, 10, 12" -> [8,10,12]\n'
                           ' "8 9, 10-12" -> [8,9,10,11,12])\n')

        group.addParam('nm_dt', params.FloatParam, label='NM time step', default=0.001,
                      help="Time step of normal modes integration. Should be equal to MD time step. Could be increase "
                           "to accelerate NM integration, however can make the simulation unstable.",
                       condition="simulationType==2 or simulationType==4",expertLevel=params.LEVEL_ADVANCED)
        group.addParam('nm_mass', params.FloatParam, default=10.0, label='NM mass',
                      help="Mass value of Normal modes for NMMD", condition="simulationType==2 or simulationType==4",
                      expertLevel=params.LEVEL_ADVANCED)
        # group.addParam('nm_init', params.FileParam, label='NM init', default=None,
        #               help="TODO", condition="simulationType==2 or simulationType==4",expertLevel=params.LEVEL_ADVANCED)
        group = form.addGroup('REMD parameters', condition="simulationType==3 or simulationType==4")
        group.addParam('exchange_period', params.IntParam, default=1000, label='Exchange Period',
                      help="Number of MD steps between replica exchanges", condition="simulationType==3 or simulationType==4")
        group.addParam('nreplica', params.IntParam, default=1, label='Number of replicas',
                      help="Number of replicas for REMD", condition="simulationType==3 or simulationType==4")

        # MD params =================================================================================================
        form.addSection(label='MD parameters')
        group = form.addGroup('Energy')
        group.addParam('implicitSolvent', params.EnumParam, label="Implicit Solvent", default=1,
                      choices=['GBSA', 'NONE'],
                      help="Turn on Generalized Born/Solvent accessible surface area model (Implicit Solvent). Boundary condition must be NO."
                           " ATDYN only.")

        group.addParam('boundary', params.EnumParam, label="Boundary", default=0,
                      choices=['No boundary', 'Periodic Boundary Condition'],
                      help="Type of boundary condition. In case of implicit solvent, "
                           " GO models or vaccum simulation, choose No boundary")
        group.addParam('box_size_x', params.FloatParam, label='Box size X',
                      help="Box size along the x dimension", condition="boundary==1")
        group.addParam('box_size_y', params.FloatParam, label='Box size Y',
                      help="Box size along the y dimension", condition="boundary==1")
        group.addParam('box_size_z', params.FloatParam, label='Box size Z',
                      help="Box size along the z dimension", condition="boundary==1")

        group.addParam('electrostatics', params.EnumParam, label="Non-bonded interactions", default=1,
                      choices=['PME', 'Cutoff'],
                      help="Type of Non-bonded interactions. "
                           " CUTOFF: Non-bonded interactions including the van der Waals interaction are just"
                           " truncated at cutoffdist; "
                           " PME : Particle mesh Ewald (PME) method is employed for long-range interactions."
                            " This option is only availabe in the periodic boundary condition")
        group.addParam('vdw_force_switch', params.BooleanParam, label="Switch function Van der Waals", default=True,
                      help="This paramter determines whether the force switch function for van der Waals interactions is"
                        " employed or not. The users must take care about this parameter, when the CHARMM"
                        " force field is used. Typically, vdw_force_switch=YES should be specified in the case of"
                        " CHARMM36",expertLevel=params.LEVEL_ADVANCED)
        group.addParam('switch_dist', params.FloatParam, default=10.0, label='Switch Distance',
                      help="Switch-on distance for nonbonded interaction energy/force quenching")
        group.addParam('cutoff_dist', params.FloatParam, default=12.0, label='Cutoff Distance',
                      help="Cut-off distance for the non-bonded interactions. This distance must be larger than"
                            " switchdist, while smaller than pairlistdist")
        group.addParam('pairlist_dist', params.FloatParam, default=15.0, label='Pairlist Distance',
                      help="Distance used to make a Verlet pair list for non-bonded interactions . This distance"
                            " must be larger than cutoffdist")

        group = form.addGroup('Ensemble', condition="simulationType!=0")
        group.addParam('ensemble', params.EnumParam, label="Ensemble", default=0,
                      choices=['NVT', 'NVE', 'NPT'],
                      help="Type of ensemble, NVE: Microcanonical ensemble, NVT: Canonical ensemble,"
                           " NPT: Isothermal-isobaric ensemble")
        group.addParam('tpcontrol', params.EnumParam, label="Thermostat/Barostat", default=1,
                      choices=['NO', 'LANGEVIN', 'BERENDSEN', 'BUSSI'],
                      help="Type of thermostat and barostat. The availabe algorithm depends on the integrator :"
                           " Leapfrog : BERENDSEN, LANGEVIN;  Velocity Verlet : BERENDSEN (NVT only), LANGEVIN, BUSSI; "
                           " NMMD : LANGEVIN (NVT only)")
        group.addParam('temperature', params.FloatParam, default=300.0, label='Temperature (K)',
                      help="Initial and target temperature")
        group.addParam('pressure', params.FloatParam, default=1.0, label='Pressure (atm)',
                      help="Target pressure in the NPT ensemble", condition="ensemble==2")

        group = form.addGroup('Contraints', condition="simulationType==1 or simulationType==3")
        group.addParam('rigid_bond', params.BooleanParam, label="Rigid bonds (SHAKE/RATTLE)",
                      default=False,
                      help="Turn on or off the SHAKE/RATTLE algorithms for covalent bonds involving hydrogen. "
                           "Must be False for NMMD.")
        group.addParam('fast_water', params.BooleanParam, label="Fast water (SETTLE)",
                      default=False,
                      help="Turn on or off the SETTLE algorithm for the constraints of the water molecules")
        group.addParam('water_model', params.StringParam, label='Water model', default="TIP3",
                      help="Residue name of the water molecule to be rigidified in the SETTLE algorithm", condition="fast_water")

        # Experiments =================================================================================================
        form.addSection(label='EM data')
        # form.addParam('EMfitChoice', params.EnumParam, label="Cryo-EM Flexible Fitting", default=0,
        #               choices=['None', 'Volume'], important=True,
        #               help="Type of cryo-EM data to be processed")
        form.addParam('EMfitChoice', params.EnumParam, label="Cryo-EM Flexible Fitting", default=0,
                      choices=['None', 'Volume (s)', 'Image (s)'], important=True,
                      help="Type of cryo-EM data to be processed")

        group = form.addGroup('Fitting parameters', condition="EMfitChoice!=0")
        group.addParam('constantK', params.StringParam, default="10000", label='Force constant (kcal/mol)',
                      help="Force constant in Eem = k*(1 - c.c.). Note that in the case of REUS, the number of "
                           " force constant value must be equal to the number of replicas, for example for 4 replicas,"
                           " a valid force constant is \"1000 2000 3000 4000\", otherwise you can specify a range of "
                           " values (for example \"1000-4000\") and the force constant values will be linearly distributed "
                           " to each replica."
                      , condition="EMfitChoice!=0")
        group.addParam('emfit_sigma', params.FloatParam, default=2.0, label="EM Fit Sigma",
                      help="Resolution parameter of the simulated map. This is usually set to the half of the resolution"
                        " of the target map. For example, if the target map resolution is 5 Å, emfit_sigma=2.5",
                      condition="EMfitChoice!=0",expertLevel=params.LEVEL_ADVANCED)
        group.addParam('emfit_tolerance', params.FloatParam, default=0.01, label='EM Fit Tolerance',
                      help="This variable determines the tail length of the Gaussian function. For example, if em-"
                        " fit_tolerance=0.001 is specified, the Gaussian function is truncated to zero when it is less"
                        " than 0.1% of the maximum value. Smaller value requires large computational cost",
                      condition="EMfitChoice!=0",expertLevel=params.LEVEL_ADVANCED)

        # Volumes
        group = form.addGroup('Volume Parameters', condition="EMfitChoice==1")
        group.addParam('inputVolume', params.PointerParam, pointerClass="Volume",
                      label="Input volume", help='Select the target EM density volume',
                      condition="EMfitChoice==1", important=True)
        group.addParam('voxel_size', params.FloatParam, default=1.0, label='Voxel size (A)',
                      help="Voxel size in ANgstrom of the target volume", condition="EMfitChoice==1")
        group.addParam('centerOrigin', params.BooleanParam, label="Center Origin", default=True,
                      help="Center the volume to the origin", condition="EMfitChoice==1")
        group.addParam('origin_x', params.FloatParam, default=0, label="Origin X",
                      help="Origin of the first voxel in X direction (in Angstrom) ",
                      condition="EMfitChoice==1 and not centerOrigin")
        group.addParam('origin_y', params.FloatParam, default=0, label="Origin Y",
                      help="Origin of the first voxel in Y direction (in Angstrom) ",
                      condition="EMfitChoice==1 and not centerOrigin")
        group.addParam('origin_z', params.FloatParam, default=0, label="Origin Z",
                      help="Origin of the first voxel in Z direction (in Angstrom) ",
                      condition="EMfitChoice==1 and not centerOrigin")

        # Images
        group = form.addGroup('Image Parameters', condition="EMfitChoice==2")
        group.addParam('inputImage', params.PointerParam, pointerClass="SetOfParticles",
                      label="Input images ", help='Select the target image set',
                      condition="EMfitChoice==2", important=True)
        group.addParam('pixel_size', params.FloatParam, default=1.0, label='Pixel size (A)',
                      help="Pixel size of the EM data in Angstrom", condition="EMfitChoice==2")
        group.addParam('projectAngleChoice', params.EnumParam, default=0, label='Projection angles',
                       choices=['same as image set', 'from xmipp file', 'from other set'],
                      help="Source of projection angles to align the input PDB with the set of images",
                       condition="EMfitChoice==2")
        group.addParam('projectAngleXmipp', params.FileParam, default=None, label='projection angle Xmipp file',
                      help="Xmipp metadata file with projection alignement parameters ",
                       condition="EMfitChoice==2 and projectAngleChoice==%i"%(PROJECTION_ANGLE_XMIPP))
        group.addParam('projectAngleImage', params.PointerParam, pointerClass="SetOfParticles",
                      label="projection angle image set  ", help='Image set containing projection alignement parameters',
                      condition="EMfitChoice==2 and projectAngleChoice==%i"%(PROJECTION_ANGLE_IMAGE))

        form.addParallelSection(threads=1, mpi=NUMBER_OF_CPU)
        # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        # Create INP files
        self._insertFunctionStep("createGenesisInputStep")

        # Convert input PDB
        self._insertFunctionStep("convertInputPDBStep")

        # Convert normal modes
        if (self.simulationType.get() == SIMULATION_NMMD or self.simulationType.get() == SIMULATION_RENMMD):
            self._insertFunctionStep("convertNormalModeFileStep")

        # Convert input EM data
        if self.EMfitChoice.get() != EMFIT_NONE:
            self._insertFunctionStep("convertInputEMStep")

        # RUN simulation
        if not self.disableParallelSim.get() and  \
            self.getNumberOfSimulation() >1  and  existsCommand("parallel") :
            self._insertFunctionStep("runSimulationParallel")
        else:
            if not self.disableParallelSim.get() and  \
                self.getNumberOfSimulation() >1  and  not existsCommand("parallel"):
                self.warning("Warning : Can not use parallel computation for GENESIS,"
                                    " please install \"GNU parallel\". Running in linear mode.")
            for i in range(self.getNumberOfSimulation()):
                inp_file = self._getExtraPath("INP_%s" % str(i + 1).zfill(6))
                outPref = self.getOutputPrefix(i)
                self._insertFunctionStep("runSimulation", inp_file, outPref)

        # Create output data
        self._insertFunctionStep("createOutputStep")

    # --------------------------- Convert Input PDBs --------------------------------------------

    def convertInputPDBStep(self):
        """
        Convert input PDB step.
        :return None:
        """

        # COPY PDBS -------------------------------------------------------------
        inputPDBfn = self.getInputPDBfn()
        n_pdb = self.getNumberOfInputPDB()
        for i in range(n_pdb):
            ext = os.path.splitext(inputPDBfn[i])[1]
            if ext == ".pdb" or ext == ".ent":
                runCommand("cp %s %s.pdb" % (inputPDBfn[i], self.getInputPDBprefix(i)))
            elif ext == ".cif" or ext == ".mmcif":
                cifToPdb(inputPDBfn[i], self.getInputPDBprefix(i)+".pdb")
            else:
                print("ERROR (toPdb), Unknown file type for file = %s" % inputPDBfn[i])

        # TOPOLOGY FILES -------------------------------------------------
        inputPrefix = self.getInputPDBprefix()
        if self.getForceField() == FORCEFIELD_CHARMM:
            if self.inputType.get() == INPUT_NEW_SIM:
                inputPSF = self.inputPSF.get()
            elif self.inputType.get() == INPUT_RESTART:
                inputPSF = self.restartProt.get().getInputPDBprefix() + ".psf"
            elif self.inputType.get() == INPUT_TOPOLOGY:
                inputPSF = self.topoProt.get()._getExtraPath("output.psf")
            runCommand("cp %s %s.psf" % (inputPSF, inputPrefix))
        elif self.getForceField() == FORCEFIELD_CAGO or self.getForceField() == FORCEFIELD_AAGO :
            if self.inputType.get() == INPUT_NEW_SIM:
                inputTOP = self.inputTOP.get()
            elif self.inputType.get() == INPUT_RESTART:
                inputTOP = self.restartProt.get().getInputPDBprefix() + ".top"
            elif self.inputType.get() == INPUT_TOPOLOGY:
                inputTOP = self.topoProt.get()._getExtraPath("output.top")
            runCommand("cp %s %s.top" % (inputTOP, inputPrefix))

        # Center PDBs -----------------------------------------------------
        if self.centerPDB.get():
            for i in range(self.getNumberOfInputPDB()):
                cmd = "xmipp_pdb_center -i %s.pdb -o %s.pdb" %\
                        (self.getInputPDBprefix(i),self.getInputPDBprefix(i))
                runCommand(cmd)
                print(cmd)

    def convertNormalModeFileStep(self):
        """
        Convert NM data step
        :return None:
        """
        nm_file = self.getInputPDBprefix() + ".nma"
        if self.modeList.empty():
            modeSelection = np.arange(7,self.inputModes.get().getSize()+1)
        else:
            modeSelection = getListFromRangeString(self.modeList.get())
        with open(nm_file, "w") as f:
            for i in range(self.inputModes.get().getSize()):
                if i+1 in modeSelection:
                    f.write(" VECTOR    %i       VALUE  0.0\n" % (i + 1))
                    f.write(" -----------------------------------\n")
                    nm_vec = np.loadtxt(self.inputModes.get()[i + 1].getModeFile())
                    for j in range(nm_vec.shape[0]):
                        f.write(" %e   %e   %e\n" % (nm_vec[j, 0], nm_vec[j, 1], nm_vec[j, 2]))

    # --------------------------- Convert Input EM data --------------------------------------------

    def convertInputEMStep(self):
        """
        Convert EM data step
        :return None:
        """
        # Convert EM data
        n_em = self.getNumberOfInputEM()
        dest_ext = "mrc" if self.EMfitChoice.get() == EMFIT_VOLUMES else "spi"
        self.readInputEMMetadata()
        inputMdName = self._getExtraPath("inputEM.xmd")
        runProgram("xmipp_image_convert", "-i %s --oext %s --oroot %s" %
                       (inputMdName, dest_ext, self._getExtraPath("inputEM_")))

        # Fix volumes origin
        if self.EMfitChoice.get() == EMFIT_VOLUMES:
            for i in range(n_em):
                # Update mrc header
                volPrefix  = self.getInputEMprefix(i)
                with mrcfile.open("%s.mrc" % volPrefix) as old_mrc:
                    with mrcfile.new("%s.mrc" % volPrefix, overwrite=True) as new_mrc:
                        new_mrc.set_data(old_mrc.data)
                        new_mrc.voxel_size = self.voxel_size.get()
                        new_mrc.header['origin'] = old_mrc.header['origin']
                        if self.centerOrigin.get():
                            origin = -np.array(old_mrc.data.shape) / 2 * self.voxel_size.get()
                            new_mrc.header['origin']['x'] = origin[0]
                            new_mrc.header['origin']['y'] = origin[1]
                            new_mrc.header['origin']['z'] = origin[2]
                        else:
                            new_mrc.header['origin']['x'] = self.origin_x.get()
                            new_mrc.header['origin']['y'] = self.origin_y.get()
                            new_mrc.header['origin']['z'] = self.origin_z.get()
                        new_mrc.update_header_from_data()
                        new_mrc.update_header_stats()

    # --------------------------- GENESIS step --------------------------------------------

    def createGenesisInputStep(self):
        """
        Create GENESIS input files
        :return None:
        """
        for indexFit in range(self.getNumberOfSimulation()):
            # INP file name
            inp_file = self._getExtraPath("INP_%s" % str(indexFit + 1).zfill(6))
            args = self.getDefaultArgs(indexFit)
            createGenesisInput(inp_file, **args)

    def getDefaultArgs(self, indexFit=0):
        inputRTF, inputPRM, inputSTR = self.getCHARMMInputs()
        args = {
            # Inputs files
            "outputPrefix": self.getOutputPrefix(indexFit),
            "inputPDBprefix": self.getInputPDBprefix(indexFit),
            "inputEMprefix": self.getInputEMprefix(indexFit),
            "rstFile": self.getRestartFile(indexFit),
            "nm_number": self.getNumberOfNormalModes(),
            "rigid_body_params": self.getRigidBodyParams(indexFit),
            "forcefield": self.getForceField(),
            "inputRTF": inputRTF,
            "inputPRM": inputPRM,
            "inputSTR": inputSTR,

            # Input Params
            "inputType": self.inputType.get(),
            "simulationType": self.simulationType.get(),
            "electrostatics": self.electrostatics.get(),
            "switch_dist": self.switch_dist.get(),
            "cutoff_dist": self.cutoff_dist.get(),
            "pairlist_dist": self.pairlist_dist.get(),
            "vdw_force_switch": self.vdw_force_switch.get(),
            "implicitSolvent": self.implicitSolvent.get(),
            "integrator": self.integrator.get(),
            "time_step": self.time_step.get(),
            "eneout_period": self.eneout_period.get(),
            "crdout_period": self.crdout_period.get(),
            "n_steps": self.n_steps.get(),
            "nbupdate_period": self.nbupdate_period.get(),
            "nm_dt": self.nm_dt.get(),
            "nm_mass": self.nm_mass.get(),
            "rigid_bond": self.rigid_bond.get(),
            "fast_water": self.fast_water.get(),
            "water_model": self.water_model.get(),
            "box_size_x": self.box_size_x.get(),
            "box_size_y": self.box_size_y.get(),
            "box_size_z": self.box_size_z.get(),
            "boundary": self.boundary.get(),
            "ensemble": self.ensemble.get(),
            "tpcontrol": self.tpcontrol.get(),
            "temperature": self.temperature.get(),
            "pressure": self.pressure.get(),
            "EMfitChoice": self.EMfitChoice.get(),
            "constantK": self.constantK.get(),
            "nreplica": self.nreplica.get(),
            "emfit_sigma": self.emfit_sigma.get(),
            "emfit_tolerance": self.emfit_tolerance.get(),
            "pixel_size": self.pixel_size.get(),
            "exchange_period": self.exchange_period.get()
        }
        return args

    def runSimulation(self, inp_file, outPref):
        """
        Run GENESIS simulations
        :return None:
        """
        programname = "atdyn" if self.md_program.get() == PROGRAM_ATDYN else "spdyn"
        params = "%s > %s.log" % (inp_file,outPref)
        env = self.getGenesisEnv()
        env.set("OMP_NUM_THREADS",str(self.numberOfThreads.get()))
        command = buildRunCommand(programname, params, numberOfMpi=self.numberOfMpi.get(),
                              hostConfig=self._stepsExecutor.hostConfig,
                              env=env)
        command = Plugin.getContinuousFlexCmd(command)
        runCommand(command, env=env)

    def runSimulationParallel(self):
        """
        Run multiple GENESIS simulations in parallel
        :return None:
        """

        # Set number of MPI per fit
        if self.getNumberOfSimulation() <= self.numberOfMpi.get():
            numberOfMpiPerFit   = self.numberOfMpi.get()//self.getNumberOfSimulation()
        else:
            if self.simulationType.get() == SIMULATION_REMD or self.simulationType.get() == SIMULATION_RENMMD:
                nreplica = self.nreplica.get()
                if nreplica > self.numberOfMpi.get():
                    raise RuntimeError("Number of MPI cores should be larger than the number of replicas.")
            else:
                nreplica = 1
            numberOfMpiPerFit   = nreplica

        # Set environnement
        env = self.getGenesisEnv()
        env.set("OMP_NUM_THREADS",str(self.numberOfThreads.get()))

        # Build command
        programname = os.path.join( Plugin.getVar("GENESIS_HOME"), "bin/atdyn")
        extradir = self._getExtraPath()
        outPath, outName = os.path.split(self.getOutputPrefix())
        outLog = os.path.join(outPath,re.sub("output_\d+", "output_{}",outName))
        params = "%s/INP_{} > %s.log " %(extradir, outLog)
        cmd = buildRunCommand(programname, params, numberOfMpi=numberOfMpiPerFit, hostConfig=self._stepsExecutor.hostConfig,
                              env=env)
        # Build parallel command
        parallel_cmd = "seq -f \"%%06g\" 1 %i | parallel -P %i \" %s\" " % (
        self.getNumberOfSimulation(),self.numberOfMpi.get()//numberOfMpiPerFit, cmd)
        parallel_cmd = Plugin.getContinuousFlexCmd(parallel_cmd)

        print("Command : %s" % cmd)
        print("Parallel Command : %s" % parallel_cmd)
        try :
            runCommand(parallel_cmd, env=env)
        except subprocess.CalledProcessError :
            print("Warning : Some processes returned with errors")



    # --------------------------- Create output step --------------------------------------------

    def createOutputStep(self):
        """
        Create output PDB or set of PDBs
        :return None:
        """
        # if self.simulationType.get() == SIMULATION_REMD or self.simulationType.get() == SIMULATION_RENMMD:
        #     self.convertReusOutputDcd()

        # Extract the pdb from the DCD file in case of SPDYN
        if self.md_program.get() == PROGRAM_SPDYN:
            for i in range(self.getNumberOfSimulation()):
                outputPrefix = self.getOutputPrefixAll(i)
                for j in outputPrefix:
                        lastPDBFromDCD(
                            inputDCD=j + ".dcd",
                            outputPDB=j + ".pdb",
                            inputPDB=self.getInputPDBprefix(i) + ".pdb")

        # In Case of CAGO, replace PDB info by input PDB because Genesis is not saving it properly
        if self.getForceField() == FORCEFIELD_CAGO:
            input = ContinuousFlexPDBHandler(self.getInputPDBprefix() + ".pdb")
            for i in range(self.getNumberOfSimulation()):
                outputPrefix = self.getOutputPrefixAll(i)
                for j in outputPrefix:
                    fn_output = j + ".pdb"
                    if os.path.exists(fn_output) and os.path.getsize(fn_output) !=0:
                        output = ContinuousFlexPDBHandler(fn_output)
                        input.coords = output.coords
                        input.write_pdb(j + ".pdb")

        # CREATE a output PDB
        if (self.simulationType.get() != SIMULATION_REMD  and self.simulationType.get() != SIMULATION_RENMMD )\
                and self.getNumberOfSimulation() == 1:
            self._defineOutputs(outputPDB=AtomStruct(self.getOutputPrefix() + ".pdb"))

        # CREATE SET OF output PDBs
        else:
            missing_pdbs = []
            pdbset = self._createSetOfPDBs("outputPDBs")
            # Add each output PDB to the Set
            for i in range(self.getNumberOfSimulation()):
                outputPrefix =self.getOutputPrefixAll(i)
                for j in outputPrefix:
                    pdb_fname = j + ".pdb"
                    if os.path.isfile(pdb_fname) and os.path.getsize(pdb_fname) != 0 :
                        pdbset.append(AtomStruct(pdb_fname))
                    else:
                        missing_pdbs.append(i+1)
            self._defineOutputs(outputPDBs=pdbset)

            # If some pdbs are missing, output a subset of the EM data that have actually been anaylzed
            if self.EMfitChoice.get() != EMFIT_NONE and len( missing_pdbs)>0:
                if self.EMfitChoice.get() == EMFIT_VOLUMES:
                    inSet = self.inputVolume.get()
                    outSet = self._createSetOfVolumes("subsetVolumes")
                    outSet.setSamplingRate(self.voxel_size.get())

                else:
                    inSet = self.inputImage.get()
                    outSet = self._createSetOfParticles("subsetParticles")
                    outSet.setSamplingRate(self.pixel_size.get())

                for i in inSet:
                    if not i.getObjId() in missing_pdbs:
                        outSet.append(i)
                self._defineOutputs(subsetEM=outSet)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = ["Genesis in a software for Molecular Dynamics Simulation, "
                   "Normal Mode Molecular Dynamics (NMMD), Replica Exchange Umbrela "
                   "Sampling (REUS) and Energy Minimization"]
        return summary

    def _validate(self):
        errors = []
        print(self.getGenesisEnv()["PATH"])
        if not os.path.exists(os.path.join(
                Plugin.getVar("GENESIS_HOME"), 'bin/atdyn')):
            errors.append("Missing GENESIS program : atdyn ")

        if not os.path.exists(os.path.join(
                Plugin.getVar("GENESIS_HOME"), 'bin/spdyn')):
            errors.append("Missing GENESIS program : spdyn ")

        return errors

    def _citations(self):
        return ["kobayashi2017genesis","vuillemot2022NMMD","harastani2022continuousflex"]

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------

    def getNumberOfInputPDB(self):
        """
        Get the number of input PDBs
        :return int: number of input PDBs
        """
        return len(self.getInputPDBfn())

    def getNumberOfInputEM(self):
        """
        Get the number of input EM data to analyze
        :return int : number of input EM data
        """
        if self.EMfitChoice.get() == EMFIT_VOLUMES:
            if isinstance(self.inputVolume.get(), SetOfVolumes): return self.inputVolume.get().getSize()
            else: return 1
        elif self.EMfitChoice.get() == EMFIT_IMAGES:
            if isinstance(self.inputImage.get(), SetOfParticles): return self.inputImage.get().getSize()
            else: return 1
        else: return 0

    def getNumberOfSimulation(self):
        """
        Get the number of simulations to perform
        :return int: Number of simulations
        """
        numberOfInputPDB = self.getNumberOfInputPDB()
        numberOfInputEM = self.getNumberOfInputEM()

        # Check input volumes/images correspond to input PDBs
        if numberOfInputPDB != numberOfInputEM and \
                numberOfInputEM != 1 and numberOfInputPDB != 1 \
                and numberOfInputEM != 0:
            raise RuntimeError("Number of input EM data and PDBs must be the same.")
        return np.max([numberOfInputEM, numberOfInputPDB])

    def getNumberOfNormalModes(self):
        if self.simulationType.get() == SIMULATION_NMMD or self.simulationType.get() == SIMULATION_RENMMD :
            if self.modeList.empty():
                modeSelection = np.arange(7,self.inputModes.get().getSize()+1)
            else:
                modeSelection\
                    = getListFromRangeString(self.modeList.get())
            return len(modeSelection)
        else:
            return None

    def getInputPDBfn(self):
        """
        Get the input PDB file names
        :return list : list of input PDB file names
        """
        initFn = []

        if self.inputType.get() == INPUT_RESTART:
            for i in range(self.restartProt.get().getNumberOfSimulation()):
                initFn += self.restartProt.get().getOutputPrefixAll(i)
            initFn = [i+".pdb" for i in initFn]
        elif self.inputType.get() == INPUT_TOPOLOGY:
            initFn = [self.topoProt.get().outputPDB.getFileName()]
        else:
            if isinstance(self.inputPDB.get(), SetOfAtomStructs) or \
                    isinstance(self.inputPDB.get(), SetOfPDBs):
                for i in range(self.inputPDB.get().getSize()):
                    initFn.append(self.inputPDB.get()[i+1].getFileName())

            else:
                initFn.append(self.inputPDB.get().getFileName())
        return initFn

    def getInputPDBprefix(self, index=0):
        """
        Get the input PDB prefix of the specified index
        :param int index: index of input PDB
        :return str: Input PDB prefix
        """
        prefix = self._getExtraPath("inputPDB_%s")
        if self.getNumberOfInputPDB() == 1:
            return prefix % str(1).zfill(6)
        else:
            return prefix % str(index + 1).zfill(6)

    def getInputEMprefix(self, index=0):
        """
        Get the input EM data prefix of the specified index
        :param int index: index of the EM data
        :return str: Input EM data prefix
        """
        prefix = self._getExtraPath("inputEM_%s")
        if self.getNumberOfInputEM() == 0:
            return ""
        elif self.getNumberOfInputEM() == 1:
            return prefix % str(1).zfill(6)
        else:
            return prefix % str(index + 1).zfill(6)


    def getOutputPrefix(self, index=0):
        """
        Output prefix of the specified index
        :param int index: index of the simulation to get
        :return string : Output prefix of the specified index
        """
        return self._getExtraPath("output_%s" % str(index + 1).zfill(6))

    def getOutputPrefixAll(self, index=0):
        """
        All output prefix of the specified index including multiple replicas in case of REUS
        :param int index: index of the simulation to get
        :return list: list of all output prefix of the specified index
        """
        outputPrefix=[]
        if self.simulationType.get() == SIMULATION_REMD or self.simulationType.get() == SIMULATION_RENMMD:
            for i in range(self.nreplica.get()):
                outputPrefix.append(self._getExtraPath("output_%s_remd%i" %
                                (str(index + 1).zfill(6), i + 1)))
        else:
            outputPrefix.append(self._getExtraPath("output_%s" % str(index + 1).zfill(6)))
        return outputPrefix

    def getRigidBodyParams(self, index=0):
        """
        Get the current rigid body parameters for the specified index in case of EMFIT with iamges
        :param int index: Index of the simulation
        :return list: angle_rot, angle_tilt, angle_psi, shift_x, shift_y
        """
        if self.EMfitChoice.get() == EMFIT_IMAGES :
            inputMd = self.getInputEMMetadata()

            idx = int(index + 1)
            params =  [
                inputMd.getValue(md.MDL_ANGLE_ROT, idx),
                inputMd.getValue(md.MDL_ANGLE_TILT, idx),
                inputMd.getValue(md.MDL_ANGLE_PSI, idx),
                inputMd.getValue(md.MDL_SHIFT_X, idx),
                inputMd.getValue(md.MDL_SHIFT_Y, idx),
            ]
            if any([i is None for i in params]):
                raise RuntimeError("Can not find angles or shifts")
            return params
        else:
            return None


    def getGenesisEnv(self):
        """
        Get environnement for running GENESIS
        :return Environ: environnement
        """
        environ = pwutils.Environ(os.environ)
        environ.set('PATH', os.path.join(Plugin.getVar("GENESIS_HOME"), 'bin'),
                    position=pwutils.Environ.BEGIN)
        return environ

    def getRestartFile(self, index=0):
        """
        Get input restart file
        :param int index: Index of the simulation
        :return str: restart file
        """
        if self.inputType.get() == INPUT_RESTART:
            if len(self.restartProt.get().getOutputPrefixAll(index))>1:
                raise RuntimeError("Multiple restart not implemented")
            rstfile = self.getInputPDBprefix(index) + ".rst"
            if not os.path.exists(rstfile):
                runCommand("cp %s.rst %s" % (self.restartProt.get().getOutputPrefix(index), rstfile))
            return rstfile
        else:
            return None

    def getForceField(self):
        """
        Get simulation forcefield
        :return int: forcefield
        """
        if self.inputType.get() == INPUT_RESTART:
            return self.restartProt.get().getForceField()
        elif self.inputType.get() == INPUT_TOPOLOGY:
            return self.topoProt.get().forcefield.get()
        else:
            return self.forcefield.get()

    def getInputEMMetadata(self):
        if self._inputEMMetadata is None:
            self._inputEMMetadata = self.readInputEMMetadata()
        return self._inputEMMetadata

    def readInputEMMetadata(self):
        nameMd = self._getExtraPath("inputEM.xmd")
        if self.EMfitChoice.get() == EMFIT_IMAGES:
            writeSetOfParticles(self.inputImage.get(), nameMd)
            inputEMMetadata = md.MetaData(nameMd)
            if self.projectAngleChoice.get() == PROJECTION_ANGLE_XMIPP:
                xmd = md.MetaData(self.projectAngleXmipp.get())
                for i in xmd:
                    rot = xmd.getValue(md.MDL_ANGLE_ROT, i)
                    tilt = xmd.getValue(md.MDL_ANGLE_TILT, i)
                    psi = xmd.getValue(md.MDL_ANGLE_PSI, i)
                    shx = xmd.getValue(md.MDL_SHIFT_X, i)
                    shy = xmd.getValue(md.MDL_SHIFT_Y, i)
                    inputEMMetadata.setValue(md.MDL_ANGLE_ROT, rot, i)
                    inputEMMetadata.setValue(md.MDL_ANGLE_TILT, tilt, i)
                    inputEMMetadata.setValue(md.MDL_ANGLE_PSI, psi, i)
                    inputEMMetadata.setValue(md.MDL_SHIFT_X, shx, i)
                    inputEMMetadata.setValue(md.MDL_SHIFT_Y, shy, i)
                inputEMMetadata.write(nameMd)
            elif self.projectAngleChoice.get() == PROJECTION_ANGLE_IMAGE:
                raise RuntimeError("projection angles from other image set error : Not implemented")

        elif self.EMfitChoice.get() == EMFIT_VOLUMES:
            if isinstance(self.inputVolume.get(), Volume):
                inputEMMetadata = md.MetaData()
                inputEMMetadata.setValue(md.MDL_IMAGE,
                                              self.inputVolume.get().getFileName(), inputEMMetadata.addObject())
                inputEMMetadata.write(nameMd)
            else:
                writeSetOfVolumes(self.inputVolume.get(), nameMd)
                inputEMMetadata = md.MetaData(nameMd)
        return inputEMMetadata

    def getCHARMMInputs(self):
        if self.forcefield.get() == FORCEFIELD_CHARMM:
            if self.inputType.get() == INPUT_RESTART:
                return self.restartProt.get().getCHARMMInputs()
            elif  self.inputType.get() == INPUT_TOPOLOGY:
                return self.topoProt.get().inputRTF.get(),self.topoProt.get().inputPRM.get(), self.topoProt.get().inputSTR.get()
            elif  self.inputType.get() == INPUT_NEW_SIM:
                return self.inputRTF.get(),self.inputPRM.get(), self.inputSTR.get()
        else:
            return None,None,None

    # def convertReusOutputDcd(self):
    #
    #     for i in range(self.getNumberOfSimulation()):
    #         remdPrefix = self._getExtraPath("output_%s_remd" % str(i + 1).zfill(6))
    #         tmpPrefix =  self._getExtraPath("output_%s_tmp" % str(i + 1).zfill(6))
    #         inp_file = self._getExtraPath("INP_tmp")
    #
    #         with open(inp_file, "w") as f:
    #             f.write("\n[INPUT]\n")
    #             f.write("reffile = %s.pdb # PDB file\n" % self.getInputPDBprefix(i))
    #             f.write("remfile = %s{}.rem  # REMD parameter ID file\n" % remdPrefix)
    #             f.write("dcdfile = %s{}.dcd  # DCD file\n" % remdPrefix)
    #             f.write("logfile = %s{}.log  # REMD energy log file\n" % remdPrefix)
    #
    #             f.write("\n[OUTPUT]\n")
    #             f.write("trjfile = %s{}.dcd  # coordinates sorted by temperature\n"% tmpPrefix)
    #             f.write("logfile = %s{}.log  # energy log sorted by temperature\n"% tmpPrefix)
    #
    #             f.write("\n[SELECTION]\n")
    #             f.write("group1 = all  # selection group 1\n")
    #
    #             f.write("\n[FITTING]\n")
    #             f.write("fitting_method = NO  # [NO,TR,TR+ROT,TR+ZROT,XYTR,XYTR+ZROT]\n")
    #             f.write("mass_weight = NO  # mass-weight is not applied\n")
    #
    #             f.write("\n[OPTION]\n")
    #             f.write("check_only = NO\n")
    #             f.write("convert_type = PARAMETER  # (REPLICA/PARAMETER)\n")
    #             f.write("num_replicas = %i  # total number of replicas used in the simulation\n"% self.nreplica.get())
    #             f.write("convert_ids =  # selected index (empty = all)(example: 1 2 5-10)\n")
    #             f.write("nsteps = %i  # nsteps in [DYNAMICS]\n" % self.n_steps.get())
    #             f.write("exchange_period = %i  # exchange_period in [REMD]\n" % self.exchange_period.get())
    #             f.write("crdout_period = %i  # crdout_period in [DYNAMICS]\n" % self.eneout_period.get() )
    #             f.write("eneout_period = %i  # eneout_period in [DYNAMICS]\n" % self.crdout_period.get() )
    #             f.write("trjout_format = DCD  # (PDB/DCD)\n")
    #             f.write("trjout_type = COOR+BOX  # (COOR/COOR+BOX)\n")
    #             f.write("trjout_atom = 1  # atom group\n")
    #             f.write("centering = NO\n")
    #             f.write("pbc_correct = NO\n")
    #
    #         runCommand("remd_convert %s"%inp_file, env=self.getGenesisEnv())
    #         for j in range(self.nreplica.get()):
    #             repPrefix = self._getExtraPath("output_%s_remd%i" % (str(i + 1).zfill(6), j+1))
    #             reptmpPrefix = self._getExtraPath("output_%s_tmp%i" % (str(i + 1).zfill(6), j+1))
    #             runCommand("mv %s.dcd %s.dcd"%(reptmpPrefix,repPrefix))
    #             runCommand("mv %s.log %s.log"%(reptmpPrefix,repPrefix))


def createGenesisInput(inp_file, outputPrefix="", inputPDBprefix="", inputEMprefix="", rstFile="", nm_number=0,
                       rigid_body_params=None, forcefield= FORCEFIELD_CAGO, inputRTF=None, inputPRM=None,
                       inputSTR=None, inputType=INPUT_NEW_SIM, simulationType=SIMULATION_MIN,
                       electrostatics=ELECTROSTATICS_CUTOFF, switch_dist=10.0, cutoff_dist=12.0,
                       pairlist_dist=15.0, vdw_force_switch=True, implicitSolvent=IMPLICIT_SOLVENT_NONE,
                       integrator=INTEGRATOR_LEAPFROG, time_step=0.001, eneout_period=100, crdout_period=100,
                       n_steps=10000, nbupdate_period=10, nm_dt=0.001, nm_mass=10.0, rigid_bond=False,
                       fast_water = False, water_model="TIP3", box_size_x=None, box_size_y=None, box_size_z=None,
                       boundary=BOUNDARY_NOBC, ensemble=ENSEMBLE_NVE, tpcontrol=TPCONTROL_NONE, temperature=300.0,
                       pressure=1.0, EMfitChoice=EMFIT_NONE, constantK=1000.0, nreplica=4, emfit_sigma=2.0,
                       emfit_tolerance=0.01, pixel_size=1.0, exchange_period=100):
    s = "\n[INPUT] \n"  # -----------------------------------------------------------
    s += "pdbfile = %s.pdb\n" % inputPDBprefix
    if forcefield == FORCEFIELD_CHARMM:
        s += "psffile = %s.psf\n" % inputPDBprefix
        s += "topfile = %s\n" % inputRTF
        s += "parfile = %s\n" % inputPRM
        if inputSTR != "" and inputSTR is not None:
            s += "strfile = %s\n" % inputSTR
    elif forcefield == FORCEFIELD_AAGO or forcefield == FORCEFIELD_CAGO:
        s += "grotopfile = %s.top\n" % inputPDBprefix
    if inputType == INPUT_RESTART:
        s += "rstfile = %s \n" % rstFile

    s += "\n[OUTPUT] \n"  # -----------------------------------------------------------
    if simulationType == SIMULATION_REMD or simulationType == SIMULATION_RENMMD:
        s += "remfile = %s_remd{}.rem\n" % outputPrefix
        s += "logfile = %s_remd{}.log\n" % outputPrefix
        s += "dcdfile = %s_remd{}.dcd\n" % outputPrefix
        s += "rstfile = %s_remd{}.rst\n" % outputPrefix
        s += "pdbfile = %s_remd{}.pdb\n" % outputPrefix
    else:
        s += "dcdfile = %s.dcd\n" % outputPrefix
        s += "rstfile = %s.rst\n" % outputPrefix
        s += "pdbfile = %s.pdb\n" % outputPrefix

    s += "\n[ENERGY] \n"  # -----------------------------------------------------------
    if forcefield == FORCEFIELD_CHARMM:
        s += "forcefield = CHARMM \n"
    elif forcefield == FORCEFIELD_AAGO:
        s += "forcefield = AAGO  \n"
    elif forcefield == FORCEFIELD_CAGO:
        s += "forcefield = CAGO  \n"

    if electrostatics == ELECTROSTATICS_CUTOFF:
        s += "electrostatic = CUTOFF  \n"
    else:
        s += "electrostatic = PME  \n"
    s += "switchdist   = %.2f \n" % switch_dist
    s += "cutoffdist   = %.2f \n" % cutoff_dist
    s += "pairlistdist = %.2f \n" % pairlist_dist
    if vdw_force_switch:
        s += "vdw_force_switch = YES \n"
    if implicitSolvent == IMPLICIT_SOLVENT_GBSA:
        s += "implicit_solvent = GBSA \n"
        s += "gbsa_eps_solvent = 78.5 \n"
        s += "gbsa_eps_solute  = 1.0 \n"
        s += "gbsa_salt_cons   = 0.2 \n"
        s += "gbsa_surf_tens   = 0.005 \n"

    if simulationType == SIMULATION_MIN:
        s += "\n[MINIMIZE]\n"  # -----------------------------------------------------------
        s += "method = SD\n"
    else:
        s += "\n[DYNAMICS] \n"  # -----------------------------------------------------------
        if simulationType == SIMULATION_NMMD or simulationType == SIMULATION_RENMMD:
            s += "integrator = NMMD  \n"
        elif integrator == INTEGRATOR_VVERLET:
            s += "integrator = VVER  \n"
        elif integrator == INTEGRATOR_LEAPFROG:
            s += "integrator = LEAP  \n"

        s += "timestep = %f \n" % time_step
    s += "nsteps = %i \n" % n_steps
    s += "eneout_period = %i \n" % eneout_period
    s += "crdout_period = %i \n" % crdout_period
    s += "rstout_period = %i \n" % n_steps
    s += "nbupdate_period = %i \n" % nbupdate_period

    if simulationType == SIMULATION_NMMD or simulationType == SIMULATION_RENMMD:
        s += "\n[NMMD] \n"  # -----------------------------------------------------------
        s += "nm_number = %i \n" % nm_number
        s += "nm_mass = %f \n" % nm_mass
        s += "nm_file = %s.nma \n" % inputPDBprefix
        # if self.nm_init.get() is not None and self.nm_init.get() != "":
        #     s += "nm_init = %s \n" % " ".join([str(i) for i in np.loadtxt(self.nm_init.get())[indexFit]])
        if nm_dt is None:
            s += "nm_dt = %f \n" % time_step
        else:
            s += "nm_dt = %f \n" % nm_dt

    if simulationType != SIMULATION_MIN:
        s += "\n[CONSTRAINTS] \n"  # -----------------------------------------------------------
        if rigid_bond:
            s += "rigid_bond = YES \n"
        else:
            s += "rigid_bond = NO \n"
        if fast_water:
            s += "fast_water = YES \n"
            s += "water_model = %s \n" % water_model
        else:
            s += "fast_water = NO \n"

    s += "\n[BOUNDARY] \n"  # -----------------------------------------------------------
    if boundary == BOUNDARY_PBC:
        s += "type = PBC \n"
        s += "box_size_x = %f \n" % box_size_x
        s += "box_size_y = %f \n" % box_size_y
        s += "box_size_z = %f \n" % box_size_z
    else:
        s += "type = NOBC \n"

    if simulationType != SIMULATION_MIN:
        s += "\n[ENSEMBLE] \n"  # -----------------------------------------------------------
        if ensemble == ENSEMBLE_NVE:
            s += "ensemble = NVE  \n"
        elif ensemble == ENSEMBLE_NPT:
            s += "ensemble = NPT  \n"
        else:
            s += "ensemble = NVT  \n"
        if tpcontrol == TPCONTROL_LANGEVIN:
            s += "tpcontrol = LANGEVIN  \n"
        elif tpcontrol == TPCONTROL_BERENDSEN:
            s += "tpcontrol = BERENDSEN  \n"
        elif tpcontrol == TPCONTROL_BUSSI:
            s += "tpcontrol = BUSSI  \n"
        else:
            s += "tpcontrol = NO  \n"
        s += "temperature = %.2f \n" % temperature
        if ensemble == ENSEMBLE_NPT:
            s += "pressure = %.2f \n" % pressure

    if (EMfitChoice == EMFIT_VOLUMES or EMfitChoice == EMFIT_IMAGES) \
            and simulationType != SIMULATION_MIN:
        s += "\n[SELECTION] \n"  # -----------------------------------------------------------
        s += "group1 = all and not hydrogen\n"

        s += "\n[RESTRAINTS] \n"  # -----------------------------------------------------------
        s += "nfunctions = 1 \n"
        s += "function1 = EM \n"
        constStr = constantK
        if "-" in constStr:
            splt = constStr.split("-")
            constStr = " ".join(
                [str(int(i)) for i in np.linspace(int(splt[0]), int(splt[1]), nreplica)])
        s += "constant1 = %s \n" % constStr
        s += "select_index1 = 1 \n"

        s += "\n[EXPERIMENTS] \n"  # -----------------------------------------------------------
        s += "emfit = YES  \n"
        s += "emfit_sigma = %.4f \n" % emfit_sigma
        s += "emfit_tolerance = %.6f \n" % emfit_tolerance
        s += "emfit_period = 1  \n"
        if EMfitChoice == EMFIT_VOLUMES:
            s += "emfit_target = %s.mrc \n" % inputEMprefix
        elif EMfitChoice == EMFIT_IMAGES:
            s += "emfit_type = IMAGE \n"
            s += "emfit_target = %s.spi \n" % inputEMprefix
            s += "emfit_pixel_size =  %f\n" % pixel_size
            s += "emfit_roll_angle = %f\n" % rigid_body_params[0]
            s += "emfit_tilt_angle = %f\n" % rigid_body_params[1]
            s += "emfit_yaw_angle =  %f\n" % rigid_body_params[2]
            s += "emfit_shift_x = %f\n" % rigid_body_params[3]
            s += "emfit_shift_y =  %f\n" % rigid_body_params[4]

        if simulationType == SIMULATION_REMD or simulationType == SIMULATION_RENMMD:
            s += "\n[REMD] \n"  # -----------------------------------------------------------
            s += "dimension = 1 \n"
            s += "exchange_period = %i \n" % exchange_period
            s += "type1 = RESTRAINT \n"
            s += "nreplica1 = %i \n" % nreplica
            s += "rest_function1 = 1 \n"

    with open(inp_file, "w") as f:
        f.write(s)


# **************************************************************************
# *
# * Authors:
# * Mohamad Harastani (mohamad.harastani@igbmc.fr)
# * Remi Vuillemot (remi.vuillemot@upmc.fr)
# * Ilyes Hamitouche (ilyes.hamitouche@upmc.fr)
# * Slavica Jonic (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es' (if scipion related)
# *  e-mail address 'slavica.jonic@upmc.fr' (for methods issues)
# **************************************************************************
import os
import pwem
from continuousflex.constants import *
import datetime
from scipion.install.funcs import VOID_TGZ
import continuousflex
import subprocess
import re
import pyworkflow.utils as pwutils


_logo = "logo.png"

MD_NMMD_GENESIS_VERSION = "1.1"
# Use this variable to activate an environment from the Scipion conda
MODEL_CONTINUOUSFLEX_ENV_ACTIVATION_VAR = "MODEL_CONTINUOUSFLEX_ENV_ACTIVATION"
# Use this general activation variable when installed outside Scipion
MODEL_CONTINUOUSFLEX_ACTIVATION_VAR = "MODEL_CONTINUOUSFLEX_ACTIVATION"

__version__ = "3.3.7"


class Plugin(pwem.Plugin):
    _homeVar = CONTINUOUSFLEX_HOME
    _pathVars = [CONTINUOUSFLEX_HOME]
    # We only support our latest release, we can't afford supporting previous releases
    _supportedVersions = [__version__]
    _url = CONTINUOUSFLEX_URL

    @classmethod
    def _defineVariables(cls):
        cls._defineVar(MODEL_CONTINUOUSFLEX_ACTIVATION_VAR, '')
        cls._defineVar(MODEL_CONTINUOUSFLEX_ENV_ACTIVATION_VAR, cls.getActivationCmd(__version__))
        cls._defineEmVar(CONTINUOUSFLEX_HOME, continuousflex.__path__[0])
        cls._defineEmVar(NMA_HOME, 'nma')
        cls._defineEmVar(GENESIS_HOME, 'MD-NMMD-Genesis-' + MD_NMMD_GENESIS_VERSION)
        cls._defineVar(VMD_HOME, '/usr/local/lib/vmd')
        cls._defineVar(MATLAB_HOME, '~/programs/Matlab')

    @classmethod
    def getEnviron(cls):
        environ = pwutils.Environ(os.environ)
        return environ

    @classmethod
    def getContinuousFlexCmd(cls, args):
        cmd = cls.getVar(MODEL_CONTINUOUSFLEX_ACTIVATION_VAR)
        if not cmd:
            cmd = cls.getCondaActivationCmd()
            cmd += cls.getVar(MODEL_CONTINUOUSFLEX_ENV_ACTIVATION_VAR)
        cmd += " && "
        cmd += args
        return cmd

    @classmethod
    def getActivationCmd(cls, version):
        return 'conda activate continuousflex-' + version

    @classmethod
    def isVersionActive(cls):
        return cls.getActiveVersion().startswith(__version__)

    @classmethod
    def getCondaLibPath(cls):
        # which python will end by /bin/python that I am replacing with /lib
        return os.popen(cls.getContinuousFlexCmd('which python')).read()[:-11] + 'lib'

    @classmethod
    def defineBinaries(cls, env):
        os.environ['PATH'] += os.pathsep + env.getBinFolder()

        def defineCondaInstallation(version):
            installed = "last-pull-%s.txt" % datetime.datetime.now().strftime("%y%h%d-%H%M%S")

            cf_commands = []
            cf_commands.append((getCondaInstallation(version, installed), installed))

            env.addPackage('ContinuousFlex', version=version,
                           commands=cf_commands,
                           tar=VOID_TGZ,
                           default=True)

            lib_path = cls.getCondaLibPath()

            env.addPackage('nma', version='3.1',
                           url='https://github.com/continuousflex-org/NMA_basic_code/raw/master/nma_v5.tar',
                           createBuildDir=False,
                           buildDir='nma',
                           target="nma",
                           commands=[('cd ElNemo; make; mv nma_* ..', 'nma_elnemo_pdbmat'),
                                     ('cd NMA_cart; LDFLAGS=-L%s make; mv nma_* ..'
                                      % lib_path, 'nma_diag_arpack')],
                           neededProgs=['gfortran'], default=True)

            target_branch = "merge_genesis_1.4"
            output = subprocess.getoutput("gfortran --version")
            gfotran_version = int(re.search(r'\d+', output).group())
            if gfotran_version >= 10:
                FFLAGS = "-fallow-argument-mismatch -ffree-line-length-none"
            else:
                FFLAGS = "-ffree-line-length-none"

            cmd = 'git clone -b %s https://github.com/continuousflex-org/MD-NMMD-Genesis.git . ; autoreconf -fi ;' \
                  ' ./configure LDFLAGS=-L\"%s\" FFLAGS=\"%s\";' \
                  ' make install;' % (target_branch, lib_path, FFLAGS)

            env.addPackage('MD-NMMD-Genesis', version=MD_NMMD_GENESIS_VERSION,
                           buildDir='MD-NMMD-Genesis', tar="void.tgz",
                           commands=[(cmd, ["bin/atdyn"])],
                           neededProgs=['mpif90'], default=True)

        def getCondaInstallation(version, txtfile):
            installationCmd = cls.getCondaActivationCmd()
            # If nvcc is not in the path, don't install Optical Flow or DeepLearning Libraries
            if os.popen('which nvcc').read() == "":
                config_path = continuousflex.__path__[0] + '/conda_noCuda.yaml'
            else:
                config_path = continuousflex.__path__[0] + '/conda.yaml'
            installationCmd += 'conda create --file {} -n continuousflex-'.format(
                config_path) + version + ' && '
            installationCmd += cls.getActivationCmd(version)
            installationCmd += ' && touch {}'.format(txtfile)
            return installationCmd

        # Install the conda environment followed by the binaries
        defineCondaInstallation(__version__)

